#ifdef GL_ES
#define LOWP lowp
#define MED mediump
#define HIGH highp
precision mediump float;
#else
#define MED
#define LOWP
#define HIGH
#endif

varying vec4 v_position;
#if defined(specularTextureFlag) || defined(specularColorFlag)
#define specularFlag
#endif

#ifdef normalFlag
varying HIGH vec3 v_normal;
#endif //normalFlag

#if defined(colorFlag)
varying vec4 v_color;
#endif

#ifdef blendedFlag
varying float v_opacity;
#ifdef alphaTestFlag
varying float v_alphaTest;
#endif //alphaTestFlag
#endif //blendedFlag

#if defined(diffuseTextureFlag) || defined(specularTextureFlag)
#define textureFlag
#endif

#ifdef diffuseTextureFlag
varying MED vec2 v_diffuseUV;
#endif

#ifdef specularTextureFlag
varying MED vec2 v_specularUV;
#endif

#ifdef diffuseColorFlag
uniform vec4 u_diffuseColor;
#endif

#ifdef diffuseTextureFlag
uniform sampler2D u_diffuseTexture;
#endif

#ifdef specularColorFlag
uniform vec4 u_specularColor;
#endif

#ifdef specularTextureFlag
uniform sampler2D u_specularTexture;
#endif

#ifdef normalTextureFlag
uniform sampler2D u_normalTexture;
#endif

#ifdef lightingFlag
varying vec3 v_lightDiffuse;

#if	defined(ambientLightFlag) || defined(ambientCubemapFlag) || defined(sphericalHarmonicsFlag)
#define ambientFlag
#endif //ambientFlag

#ifdef projectShadowFlag
#define separateAmbientFlag
#endif //projectShadowFlag

#ifdef specularFlag
varying vec3 v_lightSpecular;
#endif //specularFlag

#if defined(ambientFlag) && defined(separateAmbientFlag)
varying vec3 v_ambientLight;
#endif //separateAmbientFlag

#endif //lightingFlag

#ifdef waterFlag
uniform float u_waterLevel;
#endif //waterFlag


#ifdef projectShadowFlag
uniform sampler2D u_shadowMapTexture;
uniform vec3 u_shadowLightDirection;
uniform float u_shadowLightFar;
uniform vec3 u_shadowLightPosition;
uniform float u_shadowsEnabled;
varying vec4 v_shadowPositionLightTrans;
float  bias;

float getShadow() {
    if (u_shadowsEnabled == 0.0) {return 1.0;}
	vec3 depth = (v_shadowPositionLightTrans.xyz / v_shadowPositionLightTrans.w)*0.5+0.5;
    float visibility = 0.0;
	if ((depth.x >= 0.0) && (depth.x <= 1.0) && (depth.y >= 0.0) && (depth.y <= 1.0) ) {
        float lenToLight=dot(v_position.xyz - u_shadowLightPosition, u_shadowLightDirection)/u_shadowLightFar;
        bias = 0.00025;

		vec2 poissonPos;
		for (int a=0; a<4; a++) {
            if (a==0) poissonPos = vec2( -0.94201624, -0.39906216 );
            else if (a==1) poissonPos = vec2( 0.94558609, -0.76890725 );
            else if (a==2) poissonPos = vec2( -0.094184101, -0.92938870 );
            else  poissonPos = vec2( 0.34495938, 0.29387760 );
			vec4 lenDepthMap= texture2D(u_shadowMapTexture, depth.xy + poissonPos/2200.0);
			if((lenDepthMap.g + lenDepthMap.b/255.0) + bias < lenToLight){
				visibility+=0.25;
			}
        }

        return 1.0 - visibility;
    }
    return 1.0;
}
#endif //projectShadowFlag

#ifdef skyboxFlag
uniform sampler2D u_skyboxTexture;
uniform float u_cloudsScale;
uniform float u_cloudsHeight;
uniform float u_cloudsParallax;
uniform vec3 u_skyboxCameraPosition;
uniform vec3 u_cloudsSunDirection;
uniform vec3 u_skyColor;
varying vec4 v_skyboxFragPosition;

vec2 getCloudCover() {
  vec3 normalizedPosition = v_skyboxFragPosition.xyz - u_skyboxCameraPosition * (1.0 - u_cloudsParallax);
  if (u_cloudsHeight / normalizedPosition.y < 0.0) {
   return vec2(0.0);
  }
  vec2 uv = normalizedPosition.xz * u_cloudsScale * u_cloudsHeight / normalizedPosition.y + .5;
  float margin = .3;
  return vec2(1.0, mix(0.0, texture2D(u_skyboxTexture, uv).r, clamp((1.0 - 2.0 * length(uv - vec2(.5))), 0.0, margin) / margin));
}

vec3 getSkyColor() {
  float colorChange = .06;
  HIGH float d = - dot(10.0 * normalize(u_cloudsSunDirection), 10.0 * normalize(v_normal));
  float skyColorFactor = 1.0 - colorChange + colorChange * d / 100.0;

  vec3 sunColor = vec3(1.0);
  float sunMixThreshold1 = 0.9995;
  float sunMixEdge = 1000.0;
  HIGH float sunMix = clamp(clamp((-d / 100.0 + sunMixThreshold1) / sunMixThreshold1, 0.0, 1.0) * sunMixEdge, 0.0, 1.0);
  HIGH vec3 ret = mix(sunColor, u_skyColor * skyColorFactor, sunMix);
  return ret;
}

#endif // skyboxFlag

#ifdef cutTunnelHoleFlag
uniform sampler2D u_cutTunnelHoleTexture;
uniform int u_cutTunnelHoleEnabled;
uniform vec2 u_cutTunnelHoleTextureSize;
uniform vec3 u_cutTunnelHoleCameraPosition;
uniform vec3 u_cutTunnelHoleCameraDirection;
uniform float u_cutTunnelHoleCameraFar;
const float cutHoleScale = 2.0;
const float cutHoleCameraFar = 250.0;
#endif //cutTunnelHoleFlag


#ifdef fogFlag
uniform vec4 u_fogColor;
varying float v_fog;
#endif // fogFlag


void main() {
    #if defined(cutTunnelHoleFlag)
    if (u_cutTunnelHoleEnabled > 0) {
        vec4 cutTerrainCol = texture2D(u_cutTunnelHoleTexture, vec2(gl_FragCoord.x / u_cutTunnelHoleTextureSize.x / cutHoleScale, gl_FragCoord.y / u_cutTunnelHoleTextureSize.y / cutHoleScale));
        float lenToCamera = dot(v_position.xyz - u_cutTunnelHoleCameraPosition, u_cutTunnelHoleCameraDirection);
        float tunnelLenToCamera = (cutTerrainCol.g + cutTerrainCol.b/255.0) * cutHoleCameraFar;
        if(cutTerrainCol.r > 0.5 && tunnelLenToCamera > lenToCamera - 9.0 && tunnelLenToCamera < lenToCamera + 9.0){
            discard;
        }
    }
    #endif // cutTerrainHoleFlag

	#if defined(normalFlag)
		vec3 normal = v_normal;
	#endif // normalFlag

    #if defined(skyboxFlag)
        HIGH vec2 cloudCover = getCloudCover();
        HIGH vec3 skyColor = getSkyColor();
        HIGH vec4 diffuse = vec4(mix(skyColor, vec3(cloudCover.x), cloudCover.y), 1);
	#elif defined(diffuseTextureFlag) && defined(diffuseColorFlag) && defined(colorFlag)
		vec4 diffuse = texture2D(u_diffuseTexture, v_diffuseUV) * u_diffuseColor * v_color;
	#elif defined(diffuseTextureFlag) && defined(diffuseColorFlag)
		vec4 diffuse = texture2D(u_diffuseTexture, v_diffuseUV) * u_diffuseColor;
	#elif defined(diffuseTextureFlag) && defined(colorFlag)
		vec4 diffuse = texture2D(u_diffuseTexture, v_diffuseUV) * v_color;
	#elif defined(diffuseTextureFlag)
		vec4 diffuse = texture2D(u_diffuseTexture, v_diffuseUV);
	#elif defined(diffuseColorFlag) && defined(colorFlag)
		vec4 diffuse = u_diffuseColor * v_color;
	#elif defined(diffuseColorFlag)
		vec4 diffuse = u_diffuseColor;
	#elif defined(colorFlag)
		vec4 diffuse = v_color;
	#else
		vec4 diffuse = vec4(1.0);
	#endif



	#if (!defined(lightingFlag))
        gl_FragColor.rgb = diffuse.rgb;
	#elif (!defined(specularFlag))
		#if defined(ambientFlag) && defined(separateAmbientFlag)
			#ifdef projectShadowFlag
				gl_FragColor.rgb = (diffuse.rgb * (v_ambientLight + getShadow() * v_lightDiffuse));
			#else
				gl_FragColor.rgb = (diffuse.rgb * (v_ambientLight + v_lightDiffuse));
			#endif //projectShadowFlag
		#else
			#ifdef projectShadowFlag
				gl_FragColor.rgb = getShadow() * (diffuse.rgb * v_lightDiffuse);
			#else
				gl_FragColor.rgb = (diffuse.rgb * v_lightDiffuse);
			#endif //projectShadowFlag
		#endif
	#else
		#if defined(specularTextureFlag) && defined(specularColorFlag)
			vec3 specular = texture2D(u_specularTexture, v_specularUV).rgb * u_specularColor.rgb * v_lightSpecular;
		#elif defined(specularTextureFlag)
			vec3 specular = texture2D(u_specularTexture, v_specularUV).rgb * v_lightSpecular;
		#elif defined(specularColorFlag)
			vec3 specular = u_specularColor.rgb * v_lightSpecular;
		#else
			vec3 specular = v_lightSpecular;
		#endif

		#if defined(ambientFlag) && defined(separateAmbientFlag)
			#ifdef projectShadowFlag
                gl_FragColor.rgb = (diffuse.rgb * (getShadow() * v_lightDiffuse + v_ambientLight)) + specular;
			#else
				gl_FragColor.rgb = (diffuse.rgb * (v_lightDiffuse + v_ambientLight)) + specular;
			#endif //projectShadowFlag
		#else
			#ifdef projectShadowFlag
				gl_FragColor.rgb = getShadow() * ((diffuse.rgb * v_lightDiffuse) + specular);
			#else
				gl_FragColor.rgb = (diffuse.rgb * v_lightDiffuse) + specular;
			#endif //projectShadowFlag
		#endif
	#endif //lightingFlag

	#ifdef fogFlag
		gl_FragColor.rgb = mix(gl_FragColor.rgb, u_fogColor.rgb, v_fog);
	#endif // end fogFlag

	#ifdef blendedFlag
		gl_FragColor.a = diffuse.a * v_opacity;
		#ifdef alphaTestFlag
			if (gl_FragColor.a <= v_alphaTest)
				discard;
		#endif
	#else
		gl_FragColor.a = 1.0;
	#endif



	#if defined(waterFlag)
	    const vec4 waterColor = vec4(0.051, 0.647, 1.0, 1.0);
	    float mixAmountInitial = step(0.0, u_waterLevel - v_position.y) * (clamp(u_waterLevel - v_position.y, 0.0, 0.05)) * 20.0;
	    float mixAmountSecondary = step(0.05, u_waterLevel - v_position.y) * (clamp(u_waterLevel - v_position.y - 0.05, 0.0, 5.0)) * 0.2;
	    gl_FragColor = mix(gl_FragColor, waterColor, 0.4 * mixAmountInitial + 0.6 * mixAmountSecondary);
	#endif // waterFlag

	//float numSteps = 8.0;
	//gl_FragColor.r = floor(gl_FragColor.r * numSteps + .5)/numSteps;
	//gl_FragColor.g = floor(gl_FragColor.g * numSteps + .5)/numSteps;
	//gl_FragColor.b = floor(gl_FragColor.b * numSteps + .5)/numSteps;

}