#ifdef GL_ES
#define LOWP lowp
precision highp float;
#else
#define LOWP
#endif
varying LOWP vec4 v_color;
varying vec2 v_texCoords;
uniform sampler2D u_texture;

uniform int u_octaves;
uniform float u_lacunarity;
uniform float u_timeLacunarity;
uniform float u_time;
uniform float u_gain;
uniform float u_resolution;
uniform float u_lowThreshold;
uniform float u_highThreshold;
uniform float u_scale;
uniform vec2 u_rand;

float random(vec2 st) {
    return fract(sin(dot(st.xy, vec2(12.9898,78.233)))* 43758.5453123);
}

float noise(vec2 st) {
  vec2 i = floor(st);
  vec2 f = fract(st);
   // Four corners in 2D of a tile
  float a = random(i);
  float b = random(i + vec2(1.0, 0.0));
  float c = random(i + vec2(0.0, 1.0));
  float d = random(i + vec2(1.0, 1.0));
  vec2 u = f * f * (3.0 - 2.0 * f);

  return mix(a, b, u.x) +
          (c - a)* u.y * (1.0 - u.x) +
          (d - b) * u.x * u.y;
}

float fbm(vec2 st) {
  // Initial values
  float value = 0.0;
  float amplitude = 0.5;
  float frequency = u_scale;
  float timeFrequency = 1.0;
  for (int i = 0; i < u_octaves; i++) {
    value += amplitude * noise(st * frequency + u_time * timeFrequency);
    frequency *= u_lacunarity;
    timeFrequency *= u_timeLacunarity;
    amplitude *= u_gain;
  }
  return sqrt(value);
}

void main()
{
  vec2 st = gl_FragCoord.xy / u_resolution + u_rand;
  vec3 col = vec3(0.0);
  col.r += (fbm(st) - u_lowThreshold) / (u_highThreshold - u_lowThreshold);
  // col.gb = st;
  gl_FragColor = vec4(col, 1.0);
}

